package org.netkernel.lang.ncode;

import java.lang.reflect.Method;
import java.util.*;
import java.util.concurrent.locks.ReentrantLock;

import org.netkernel.ext.system.representation.IRepDeployedModules;
import org.netkernel.lang.dpml.representation.DPMLRepresentation;
import org.netkernel.layer0.bnf.BNFGrammarFactory;
import org.netkernel.layer0.bnf.IGrammarGroup;
import org.netkernel.layer0.meta.IEndpointArgumentMeta;
import org.netkernel.layer0.meta.IEndpointMeta;
import org.netkernel.layer0.meta.ISpaceElements;
import org.netkernel.layer0.meta.impl.EndpointMetaBuilder;
import org.netkernel.layer0.meta.impl.MetadataUtils;
import org.netkernel.layer0.meta.impl.SpaceElementsImpl;
import org.netkernel.layer0.nkf.*;
import org.netkernel.layer0.nkf.impl.NKFRequestImpl;
import org.netkernel.layer0.nkf.impl.NKFResponseReadOnlyImpl;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.IHDSNodeList;
import org.netkernel.layer0.representation.impl.HDSBuilder;
import org.netkernel.layer0.urii.SimpleIdentifierImpl;
import org.netkernel.layer0.urii.ValueSpace;
import org.netkernel.layer0.util.PairList;
import org.netkernel.layer0.util.TupleList;
import org.netkernel.layer0.util.Utils;
import org.netkernel.module.standard.endpoint.ConfiguredEndpointImpl.IConfig;
import org.netkernel.module.standard.endpoint.ConfiguredEndpointImpl.NullConfig;
import org.netkernel.request.IRequestResponseFields;
import org.netkernel.urii.IIdentifier;
import org.netkernel.urii.IMetaRepresentation;
import org.netkernel.urii.ISpace;

public class NCODEConfig implements IConfig
{
	private static final PairList VALIDATION = new PairList(4);
	static
	{	VALIDATION.put("/*[name()!='flows']", "MSG_VHOST_V1");
		VALIDATION.put("/flows/*[name()!='flow']", "MSG_VHOST_V2");
		VALIDATION.put("/flows/flow/*[name()!='id' and name()!='name' and name()!='desc' and name()!='grammar' and name()!='verbs']", "MSG_VHOST_V3");
		VALIDATION.put("/flows/flow[not(id) or not(name) or not(grammar)]", "MSG_VHOST_V4");
	}	
	
	private final ReentrantLock mGenerateDPMLLock = new ReentrantLock();
	private final NCODEPrototype mEndpoint;
	private final Map<String,Flow> mFlowsById;
	private INKFResponseReadOnly mDPMLScriptResponse;
	private ISpace mProtoSpace;
	private IHDSNode mConfig;
	
	private IIdentifier mDynamicImportConfigIdentifier;
	private IEndpointMeta mDynamicImportConfigMeta;
	private IHDSNode mDynamicImportConfigRep;
	
	public NCODEConfig(IHDSNode aConfig, NCODEPrototype aEndpoint, INKFRequestContext aContext, String aDynamicImportConfig) throws Exception
	{	mEndpoint=aEndpoint;
		mConfig=aConfig;
		
		//no validation on the assumption that state won't be human edited
		/*
		PairList validation=FastSchematron.validate(aConfig.getDocumentElement(), VALIDATION);
		if (validation.size()>0)
		{	String m = FastSchematron.formatValidationError(validation,aContext);			
			throw aContext.createFormattedException("EX_DF_VALIDATE", "MSG_RAW", null,null, m);
		}
		XMLReadable r = new XMLReadable(aConfig.getDocumentElement());
		List<Node> flows=(List<Node>)r.getNodes("flow");
		*/
		
		mFlowsById = new LinkedHashMap<String,Flow>();
		for (IHDSNode flowNode: aConfig.getNodes("/cde/flow__A/flow"))
		{	Flow flow=new Flow(flowNode,aContext);
			mFlowsById.put(flow.getId(),flow);
		}
		
		if (aDynamicImportConfig!=null)
		{	mDynamicImportConfigIdentifier=new SimpleIdentifierImpl(Utils.generateUniqueEndpointIdentifier(aEndpoint.getSpace(), aEndpoint, null));
			EndpointMetaBuilder emb=new EndpointMetaBuilder();
			IGrammarGroup g =  BNFGrammarFactory.createGroup(null);
			BNFGrammarFactory.createText(g, aDynamicImportConfig);
			emb.setGrammarRoot(g);
			emb.setName("dynamicImportConfig");
			emb.setVerbs(INKFRequestReadOnly.VERB_SOURCE);
			mDynamicImportConfigMeta=emb.getMeta(aContext.getKernelContext().getKernel().getLogger());
			
			IHDSNodeList imports=aConfig.getNodes("/cde/settings/dynamicImport__A/dynamicImport");
		    HDSBuilder b = new HDSBuilder();
		    b.pushNode("config");
			for (IHDSNode importNode: imports)
			{	String id=(String)importNode.getFirstValue("id");
				b.pushNode("space");
				b.addNode("uri",id);
				String version=(String)importNode.getFirstValue("version");
				if (version!=null)
				{	b.addNode("version",version);
				}
				b.popNode();
			}
			mDynamicImportConfigRep=b.getRoot();
		}
		
	}
	
	Map<String,Flow> getFlowsById()
	{	return mFlowsById;
	}
	
	IHDSNode getRawConfig()
	{	return mConfig;
	}
	
	public boolean supportsDynamicImports()
	{	return mDynamicImportConfigIdentifier!=null;
	}
	
	public IMetaRepresentation getElementMeta(String aElementId)
	{
		IMetaRepresentation result;
		Flow flow=mFlowsById.get(aElementId);
		if (flow!=null)
		{	result=flow.getMeta();
		}
		else if (aElementId.equals(mDynamicImportConfigIdentifier.toString()))
		{	result=mDynamicImportConfigMeta;
		}
		else
		{	result=null;
		}
		return result;

	}

	public ISpaceElements getElements(INKFRequestContext aContext) throws Exception
	{	TupleList tl=new TupleList(3, mFlowsById.size()+1);
		for (Flow flow: mFlowsById.values())
		{	tl.put(new SimpleIdentifierImpl(flow.getId()),null,mEndpoint);
		}
		if (mDynamicImportConfigIdentifier!=null)
		{	tl.put(mDynamicImportConfigIdentifier,null,mEndpoint);
		}
		return new SpaceElementsImpl(tl);
	}

	@Override
	public String getResolvedElement(INKFResolutionContext aContext) throws Exception
	{	String result=null;
		INKFRequestReadOnly toResolve=aContext.getRequestToResolve();
		for (Flow flow : mFlowsById.values())
		{	IEndpointMeta meta = flow.getMeta();
			if (MetadataUtils.match((IEndpointMeta)meta, toResolve, aContext))
			{	result = flow.getId();
				break;
			}
		}
		if (result==null && mDynamicImportConfigMeta!=null && MetadataUtils.match(mDynamicImportConfigMeta,toResolve, aContext))
		{	result=mDynamicImportConfigIdentifier.toString();
		}
		return result;
	}

	@Override
	public ISpace[] getSpaces(INKFRequestContext aContext) throws Exception
	{	return NullConfig.NO_SPACES;
	}

	@Override
	public void onRequest(String aElementId, INKFRequestContext aContext) throws Exception
	{	if (mDynamicImportConfigIdentifier!=null && aElementId.equals(mDynamicImportConfigIdentifier.toString()))
		{	onDynamicImportConfig(aContext);
		}
		else
		{	onRegularRequest(aElementId,aContext);
		}
	}
	
	public void onDynamicImportConfig(INKFRequestContext aContext) throws Exception
	{	aContext.createResponseFrom(mDynamicImportConfigRep);
	}

	public void onRegularRequest(String aElementId, INKFRequestContext aContext) throws Exception
	{	
		mGenerateDPMLLock.lock();
		try
		{	if (mDPMLScriptResponse==null)
			{	
				Object[] oa=getSpaceForPrototypeInstance(aContext);
				mProtoSpace=(ISpace)oa[0];
				String entity=(String)oa[1];
				// create DPML Script
				INKFRequest sreq=aContext.createRequest("active:ToDPML");
				sreq.injectRequestScope(mProtoSpace);
				sreq.addArgumentByValue("config", this);
				sreq.addArgumentByValue("entity", entity);
				//sreq.setRepresentationClass(DPMLRepresentation.class);
				mDPMLScriptResponse=aContext.issueRequestForResponse(sreq);
			}
		}
		finally
		{	mGenerateDPMLLock.unlock();
		}
		
		Flow flow=mFlowsById.get(aElementId);
		IEndpointMeta flowMeta=flow.getMeta();
		
		//create request to execute DPML script
		INKFRequestReadOnly reqIn=aContext.getThisRequest();
		NKFRequestImpl req= (NKFRequestImpl)aContext.createRequest("active:dpml");
		req.injectRequestScope(mProtoSpace);
		req.setVerb(reqIn.getVerb());
		req.setRepresentationClass(reqIn.getRepresentationClass());
		INKFResponseReadOnly primaryResponse = reqIn.getPrimaryAsResponse();
		if (primaryResponse!=null)
		{	req.addPrimaryArgumentFromResponse(primaryResponse);
		}
		IRequestResponseFields headers=reqIn.getHeaders();
		for (int i=0; i<headers.size(); i++)
		{	req.setHeader(headers.getKey(i), headers.getValue(i), headers.isSticky(i));
		}
		for (int i=0; i<reqIn.getArgumentCount(); i++)
		{	String argName=reqIn.getArgumentName(i);
			String argValue=reqIn.getArgumentValue(i);
			if (org.netkernel.lang.ncode.Utils.shouldConvertArgumentToString(flowMeta,argName))
			{	req.addArgumentByValue(argName, argValue);
			}
			else
			{	req.addArgumentAsAbsolute(argName, argValue);
			}
		}
		
		ValueSpace valueSpace=new ValueSpace(1);
		String scriptId=mEndpoint.getParameter("operator").toString();
		//valueSpace.put(scriptId, mDPMLScriptResponse.getRepresentation(), ((NKFResponseReadOnlyImpl)mDPMLScriptResponse).getMeta());
		//TODO fix dpml script response expiry passing so we can ignore module list but still keep dependencies on other stuff
		valueSpace.put(scriptId, mDPMLScriptResponse.getRepresentation(), null);
		req.addArgument("operator", scriptId);
		req.addArgumentAsAbsolute("response", aElementId);
		ISpace dpmlSpace=aContext.getKernelContext().getKernel().getSpace(new SimpleIdentifierImpl("urn:org:netkernel:lang:dpml"), null, null);
		req.injectRequestScope(dpmlSpace);
		req.injectDurableRequestScope(valueSpace);
		

		/*
		req.addArgumentFromResponse("operator", mDPMLScriptResponse);
		req.addArgumentAsAbsolute("response", aElementId);
		ISpace dpmlSpace=aContext.getKernelContext().getKernel().getSpace(new SimpleIdentifierImpl("urn:org:netkernel:lang:dpml"), null, null);
		req.injectRequestScope(dpmlSpace);
		*/
		
		req.setRepresentationClass(reqIn.getRepresentationClass());
		INKFResponseReadOnly dpmlExecResponse=aContext.issueRequestForResponse(req);
		aContext.createResponseFrom(dpmlExecResponse);
	}
	
	NCODEPrototype getOwningEndpoint()
	{	return mEndpoint;
	}
	
	/** Determine the inner nCoDE space where the prototype is declared. This is useful as the instance
	 * needs to inject into this space.
	 */
	private Object[] getSpaceForPrototypeInstance(INKFRequestContext aContext) throws Exception
	{	IIdentifier spaceId=new SimpleIdentifierImpl("urn:org:netkernel:ext:system");
		ISpace runtimeSpace=aContext.getKernelContext().getKernel().getSpace(spaceId, null, null);
		INKFRequest mreq=aContext.createRequest("active:moduleListProtoMap");
		mreq.injectRequestScope(runtimeSpace);
		Object o=aContext.issueRequest(mreq);
		Class c=o.getClass();
		Method m=c.getMethod("getPrototypeHashFor", Class.class);
		String protoHash=(String)m.invoke(o, mEndpoint.getClass());
		int i=protoHash.lastIndexOf('/');
		String protoSpaceHash=protoHash.substring(0,i);
		mreq=aContext.createRequest("active:moduleList");
		mreq.injectRequestScope(runtimeSpace);
		o=aContext.issueRequest(mreq);
		c=o.getClass();
		m=c.getMethod("componentForHash", String.class);
		ISpace space= (ISpace)m.invoke(o, protoSpaceHash);
		
		m=c.getMethod("hashForComponent", Object.class);
		String entity= (String)m.invoke(o, mEndpoint);
		
		return new Object[] { space,entity };
	}

}
