package org.netkernel.lang.ncode;

import org.netkernel.layer0.meta.IEndpointArgumentMeta;
import org.netkernel.layer0.nkf.INKFLocale;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.NKFException;
import org.netkernel.layer0.representation.IBinaryStreamRepresentation;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.impl.HDSBuilder;
import org.netkernel.layer0.util.OrderedCheapMultiStringMap;
import java.util.*;

public class PaletteRepresentation
{
	private Map<String,IPaletteItem> mItemsById = new HashMap<String, PaletteRepresentation.IPaletteItem>();
	private Map<String,PaletteGroup> mGroupsById = new HashMap<String, PaletteRepresentation.PaletteGroup>();
	private PaletteGroup mOrphanGroup;
	
	public PaletteRepresentation()
	{	mOrphanGroup = new PaletteGroup("orphaned","Ungrouped","this group contains all endpoints without groups defined",Integer.MAX_VALUE);
		mGroupsById.put(mOrphanGroup.mId, mOrphanGroup);
	}
	
	public PaletteGroup addGroup(String aId, String aName, String aDesc, int aOrder)
	{	PaletteGroup g = new PaletteGroup(aId,aName,aDesc,aOrder);
		mGroupsById.put(aId, g);
		return g;
	}
	
	public void addEndpoint(String aPaletteId, IPaletteItem aItem, INKFRequestContext aContext) throws NKFException
	{	PaletteGroup group = mGroupsById.get(aPaletteId);
		if (group==null)
		{	group=mOrphanGroup;
			//aContext.logFormatted(INKFLocale.LEVEL_WARNING, "MSG_NCODE_PALETTE_NO_GROUP", aPaletteId, aItem.getPaletteId());
		}
	
		group.addItem(aItem);
		mItemsById.put(aItem.getPaletteId(), aItem);
		//throw aContext.createFormattedException("EX_NCODE_RUNTIME", "MSG_NCODE_PALETTE_NO_GROUP", null, null, aPaletteId,aItem.getPaletteId());
	}
	
	public IHDSNode getPaletteAsHDS()
	{
		HDSBuilder b = new HDSBuilder();
		List<PaletteGroup> groups=new ArrayList<PaletteGroup>(mGroupsById.values());
		Collections.sort(groups);
		b.pushNode("group__A");
		for (PaletteGroup group : groups)
		{	if (!group.isEmpty())
			{	group.serializeTo(b);
			}
		}
		b.popNode();
		return b.getRoot();
	}
	
	public IPaletteItem getPaletteItem(String aId, INKFRequestContext aContext) throws NKFException
	{	IPaletteItem result=mItemsById.get(aId);
		if (result==null)
		{	throw aContext.createFormattedException("EX_NCODE_RUNTIME", "MSG_NCODE_PALETTE_NO_ITEM", null, null, aId);
		}
		return result;
	}
	
	private static class PaletteGroup implements Comparable<PaletteGroup>
	{
		private final String mId;
		private final String mName;
		private final String mDesc;
		private final int mOrder;
		private final List<IPaletteItem> mItems;
		
		public PaletteGroup(String aId, String aName, String aDesc, int aOrder)
		{	mId=aId;
			mName=aName;
			mDesc=aDesc;
			mOrder=aOrder;
			mItems=new ArrayList<PaletteRepresentation.IPaletteItem>();
		}
		
		public void addItem(IPaletteItem aItem)
		{	mItems.add(aItem);
			Collections.sort(mItems);
		}
		
		public boolean isEmpty()
		{	return mItems.size()==0;
		}
		
		@Override
		public int compareTo(PaletteGroup o)
		{	return mOrder-o.mOrder;
		}
		
		public void serializeTo(HDSBuilder b)
		{	b.pushNode("group");
			b.addNode("id", mId);
			b.addNode("name", mName);
			b.addNode("desc", mDesc);
			b.pushNode("endpoint__A");
			List<IPaletteItem> items=new ArrayList<IPaletteItem>(mItems);
			Collections.sort(items);
			for (IPaletteItem item: items)
			{	item.serializeTo(b);
			}
			b.popNode();
			b.popNode();
		}
	}
	
	public interface IPaletteItem extends Comparable<IPaletteItem>
	{
		int getOrder();
		String getEndpointId();
		String getPaletteId();
		void serializeTo(HDSBuilder aBuilder);
		IHDSNode validateInputs(IHDSNode aInputs, INKFRequestContext aContext) throws Exception;
		IBinaryStreamRepresentation getIcon(INKFRequestContext aContext) throws Exception;
		IHDSNode validateValue(String aValue);
		OrderedCheapMultiStringMap getHardWiredInputs();
		IEndpointArgumentMeta.ArgumentType getInputType(String aInputName);
		String getDocumentation();
	}
}