package org.netkernel.lang.ncode.editor;

import java.io.*;
import java.util.regex.Pattern;

import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.impl.HDSBuilder;
import org.netkernel.layer0.util.FastSchematron;
import org.netkernel.layer0.util.PairList;
import org.netkernel.xml.util.XMLUtils;
import org.w3c.dom.Document;

public class ValueValidatorFactory
{
	public final IValueValidator mNoValidation;
	public final IValueValidator mXMLValidation;
	public final IValueValidator mHDSValidation;
	public final IValueValidator mURIValidation;
	public final IValueValidator mBooleanValidation;
	public final IValueValidator mIntegerValidation;
	public final IValueValidator mFloatValidation;
	public final IValueValidator mPOJOValidation;
	
	private static final ValueValidatorFactory sInstance=new ValueValidatorFactory();
	
	private static final PairList POJO_VALIDATION = new PairList(5);
	static
	{	
		POJO_VALIDATION.put("/*[name()!='literal']","Root element must be literal");
		POJO_VALIDATION.put("/*[not(@type)]","Literal must have type attribute");
		POJO_VALIDATION.put("/*/*[name()!='literal']","Constructor parameters must be nested literals");
		POJO_VALIDATION.put("/*/*[not(@type)]","Literal must have type attribute");
	}
	
	
	
	public static ValueValidatorFactory getInstance()
	{	return sInstance;
	}
	
	private static class RegexValidator extends IValueValidator
	{
		private final Pattern mPattern;
		private final String mName;
		
		public RegexValidator(String aName, String aPattern)
		{	mName=aName;
			mPattern=Pattern.compile(aPattern);
		}
		public IHDSNode validate(String aValue)
		{	if (mPattern.matcher(aValue).matches())
			{	return IValueValidator.VALID;
			}
			else
			{	HDSBuilder b=new HDSBuilder();
				b.addNode("valid", false);
				String m="Malformed "+mName+", must match regex "+mPattern.pattern();
				b.addNode("message", m);
				return b.getRoot();
			}
		}
	}
	
	private static IHDSNode createValidationError(String aMessage)
	{	HDSBuilder b=new HDSBuilder();
		b.addNode("valid", false);
		b.addNode("message", aMessage);
		return b.getRoot();
	}
	
	
	public ValueValidatorFactory()
	{
		
		mNoValidation=new IValueValidator()
		{	public IHDSNode validate(String aValue)
			{	return IValueValidator.VALID;
			}
		};
		
		mXMLValidation=new IValueValidator()
		{
			public IHDSNode validate(String aValue)
			{	try
				{	Reader reader=new StringReader(aValue);
					Document result=XMLUtils.parse(reader);
					return IValueValidator.VALID;
				}
				catch (Exception e)
				{	HDSBuilder b=new HDSBuilder();
					b.addNode("valid", false);
					b.addNode("message", e.getMessage());
					return b.getRoot();
				}
			}
		};
		
		mHDSValidation=new IValueValidator()
		{
			public IHDSNode validate(String aValue)
			{	try
				{	Reader reader=new StringReader("<hds>"+aValue+"</hds>");
					Document result=XMLUtils.parse(reader);
					if (XMLUtils.getFirstChildElement(result.getDocumentElement())==null)
					{	throw new Exception("expected at least one node");
					}
					if (XMLUtils.getText(result.getDocumentElement()).trim().length()>0)
					{	throw new Exception("no text expected outside node");
					}
					return IValueValidator.VALID;
				}
				catch (Exception e)
				{	HDSBuilder b=new HDSBuilder();
					b.addNode("valid", false);
					b.addNode("message", e.getMessage());
					return b.getRoot();
				}
			}
		};
		
		
		mPOJOValidation=new IValueValidator()
		{
			public IHDSNode validate(String aValue)
			{	try
				{	Reader reader=new StringReader(aValue);
					Document result=XMLUtils.parse(reader);
					PairList validation=FastSchematron.validate(result.getDocumentElement(), POJO_VALIDATION);
					if (validation.size()>0)
					{	StringBuilder sb=new StringBuilder();
						for (int i=0; i<validation.size(); i++)
						{	if (i>0)
							{	sb.append(", ");
							}
							String errorKey = (String)validation.getValue2(i);
							sb.append(errorKey);
						}
						return createValidationError(sb.toString());
					}					
					return IValueValidator.VALID;
				}
				catch (Exception e)
				{	HDSBuilder b=new HDSBuilder();
					b.addNode("valid", false);
					b.addNode("message", e.getMessage());
					return b.getRoot();
				}
			}
		};
		
		
		mURIValidation=new RegexValidator("URI","^(?:([A-Za-z][A-Za-z\\+-]+):)?(\\/{0,3})(([0-9.\\-A-Za-z]+)(?::(\\d+))?)?(?:\\/([^?#]*))?(?:\\?([^#]*))?(?:#(.*))?$");
		mBooleanValidation=new RegexValidator("Boolean","TRUE|true|FALSE|false");
		mIntegerValidation=new RegexValidator("Integer","[\\+-]?\\d+");
		mFloatValidation=new RegexValidator("Float","[-\\+]?[0-9]*\\.?[0-9]+([eE][-\\+]?[0-9]+)?");
	}	
}
